class PanicException(Exception):
    def __init__(self, place: str, reason: str, suggestion: str):
        super().__init__("\n".join([place, reason, suggestion]))
        self.place: str = place
        self.reason: str = reason
        self.suggestion: str = suggestion

    def print(self) -> None:
        print('🔥 ' + self.place)
        print(self.reason)
        if self.suggestion:
            print(self.suggestion)



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import ListRead

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_LR(state: State, arg1: str, arg2: str, arg3: str) -> None:
    state.require(state.in_action, 'Not inside any action.', 'LR can be used inside actions.')
    dst = Argument(state, arg1)
    list_ = Argument(state, arg2)
    idx = Argument(state, arg3)
    state.require(
        dst.list_read_context(list_, idx), 
        'Mismatched types in the list read context.', 
        f'ARG1 {dst.explain()}, ARG2 {list_.explain()}, ARG3 {idx.explain()}'
    )
    
    state.last_action.add_instruction(ListRead(dst, list_, idx))



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aasm.intermediate.argument import Argument


class Declaration:  
    def __init__(self, name: Argument, value: Argument):
        self.name: str = name.expr
        self.value: Argument = value
        
    def print(self) -> None:
        print(f'Declaration: {self.name}')
        self.value.print()



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import ExpDist, NormalDist, UniformDist
from aasm.parsing.op.round import op_ROUND

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_RAND(state: State, arg1: str, arg2: str, arg3: str, args: List[str]) -> None:
    state.require(state.in_action, 'Not inside any action.', f'RAND can be used inside actions.')
    result = Argument(state, arg1)

    match arg3, args:
        case 'uniform', [ a, b ]:
            a_arg = Argument(state, a)
            b_arg = Argument(state, b)
            state.require(
                result.random_number_generation_context(a_arg, b_arg), 
                'Mismatched types in the random number generation context.', 
                f'RESULT {result.explain()}, A {a_arg.explain()}, B {b_arg.explain()}'
            )

            state.last_action.add_instruction(UniformDist(result, a_arg, b_arg))

        case 'normal', [ mean, std_dev ]:
            mean_arg = Argument(state, mean)
            std_dev_arg = Argument(state, std_dev)
            state.require(
                result.random_number_generation_context(mean_arg, std_dev_arg), 
                'Mismatched types in the random number generation context.', 
                f'RESULT {result.explain()}, MEAN {mean_arg.explain()}, STD_DEV {std_dev_arg.explain()}'
            )

            state.last_action.add_instruction(NormalDist(result, mean_arg, std_dev_arg))

        case 'exp', [ lambda_ ]:
            lambda_arg = Argument(state, lambda_)
            state.require(
                result.random_number_generation_context(lambda_arg), 
                'Mismatched types in the random number generation context.', 
                f'RESULT {result.explain()}, LAMBDA {lambda_arg.explain()}'
            )

            state.last_action.add_instruction(ExpDist(result, lambda_arg))

        case _:
            state.panic(f'Incorrect operation: RAND {arg1} {arg2} {arg3} {args}')

    match arg2:
        case 'float':
            ...

        case 'int':
            op_ROUND(state, arg1)

        case _:
            state.panic(f'Incorrect operation: RAND {arg1} {arg2} {arg3} {args}')



import keyword
from decimal import Decimal
from typing import List

from aasm.utils.iteration import zip_consecutive_pairs


def is_float(value: str) -> bool:
    try:
        float(value)
        return True
    except ValueError:
        return False


def is_int(value: str) -> bool:
    return value.isdigit() or (value[0] == '-' and value[1:].isdigit())


def is_valid_enum_list(enums: List[str]) -> bool:
    if len(enums) == 0 or len(enums) % 2:
        return False
    total_sum = Decimal(0.0)
    for enum_pair in zip_consecutive_pairs(enums):
        if not is_float(enum_pair[1]) or float(enum_pair[1]) < 0.0:
            return False
        total_sum += Decimal(float(enum_pair[1]))
    if total_sum < 99.0 or total_sum > 101.0:
        return False
    return True


def is_valid_name(name: str) -> bool:
    return (len(name) != 0 
            and not name[0].isdigit() 
            and (name.isalnum() or "_" in name) 
            and name.lower() not in get_invalid_names())


def get_invalid_names() -> List[str]:
    invalid_names = [ 'send', 'rcv', 'len', 'round', 'list', 'filter', 'self', 'jid',
                      'datetime', 'random', 'numpy', 'json', 'spade', 'copy', 'uuid',
                      'get_json_from_spade_message', 'get_spade_message', 'logger',
                      'any', 'sys', 'limit_number', 'int', 'BackupBehaviour', 
                      'backup_url', 'backup_period', 'backup_delay' ]
    invalid_names.extend(keyword.kwlist)
    return invalid_names


def print_invalid_names() -> str:
    return ", ".join(get_invalid_names())



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.generating.python_code import PythonCode
from aasm.intermediate.graph import (AgentConstantAmount, AgentPercentAmount,
                                     ConnectionConstantAmount,
                                     ConnectionDistExpAmount,
                                     ConnectionDistNormalAmount,
                                     ConnectionDistUniformAmount,
                                     StatisticalGraph)

if TYPE_CHECKING:
    from aasm.intermediate.graph import Graph


class PythonGraph(PythonCode):
    def __init__(self, indent_size: int, graph: Graph | None):
        super().__init__(indent_size)
        if graph:
            self.add_required_imports()
            self.add_newlines(2)
            self.generate_graph(graph)

    def add_required_imports(self) -> None:
        self.add_line('import random')
        self.add_line('import uuid')
        self.add_line('import numpy')
        
    def generate_graph(self, graph: Graph) -> None:
        match graph:
            case StatisticalGraph():
                self.add_statistical_graph(graph)
                
            case _:
                raise Exception(f'Unknown graph type: {graph.print()}')            

    def add_statistical_graph(self, graph: StatisticalGraph) -> None:
        self.add_line('def generate_graph_structure(domain):')
        self.indent_right()

        if not graph.agents:
            self.add_line('return []')
            self.indent_left()
            return
        
        num_agents_expr: List[str] = []
        for agent in graph.agents.values():                
            match agent.amount:
                case AgentConstantAmount():
                    self.add_line(f'_num_{agent.name} = {agent.amount.value}')
                    
                case AgentPercentAmount():
                    self.add_line(f'_num_{agent.name} = round({agent.amount.value} / 100 * {graph.size})')
                    
                case _:
                    raise Exception(f'Unknown agent amount type: {agent.amount.print()}')
                
            num_agents_expr.append(f'_num_{agent.name}')
        
        self.add_line(f'num_agents = {" + ".join(num_agents_expr)}')
        self.add_line('random_id = str(uuid.uuid4())[:5]')
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]')
        self.add_line('agents = []')
        self.add_line('next_agent_idx = 0')
        for agent in graph.agents.values():
            self.add_line(f'for _ in range(_num_{agent.name}):')
            self.indent_right()
            
            match agent.connections:
                case ConnectionConstantAmount():
                    self.add_line(f'num_connections = {agent.connections.value}')
                
                case ConnectionDistNormalAmount():
                    self.add_line(f'num_connections = int(numpy.random.normal({agent.connections.mean}, {agent.connections.std_dev}))')
                    
                case ConnectionDistExpAmount():
                    self.add_line(f'num_connections = int(numpy.random.exponential(1 / {agent.connections.lambda_}))')
                    
                case ConnectionDistUniformAmount():
                    self.add_line(f'num_connections = int(random.uniform({agent.connections.a}, {agent.connections.b}))')
                    
                case _:
                    raise Exception(f'Unknown connection amount: {agent.connections.print()}')
            
            self.add_line('num_connections = max(min(num_connections, len(jids) - 1), 0)')
            self.add_line('jid = jids[next_agent_idx]')
            self.add_line('agents.append({')
            self.indent_right()
            self.add_line('"jid": jid,')
            self.add_line(f'"type": "{agent.name}",')
            self.add_line('"connections": random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections),')
            self.indent_left()
            self.add_line('})')
            self.add_line('next_agent_idx += 1')
            self.indent_left()
        
        self.add_line('return agents')
        self.indent_left()



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.graph import (AgentConstantAmount, AgentPercentAmount,
                                     ConnectionConstantAmount,
                                     ConnectionDistExpAmount,
                                     ConnectionDistNormalAmount,
                                     ConnectionDistUniformAmount,
                                     StatisticalAgent, StatisticalGraph)
from aasm.utils.validation import is_float, is_int

if TYPE_CHECKING:
    from parsing.state import State

    from aasm.intermediate.graph import AgentAmount, ConnectionAmount


def op_DEFG(state: State, agent_name: str, amount: str, args: List[str]) -> None:    
    state.require(
        state.in_graph, 
        'Cannot define agent graph amount outside graph scope.', 
        'Try defining new graphs using GRAPH.'
    )
    state.require(
        isinstance(state.last_graph, StatisticalGraph), 
        'DEFG can be used with statistical graphs.', 
        'Define statistical graphs with GRAPH statistical.'
    )
    state.require(state.agent_exists(agent_name), f'Agent {agent_name} is not defined.')
    
    agent_amount: AgentAmount | None = None
    if amount.endswith('%'):
        percent_value = amount[:-1]
        state.require(is_float(percent_value), f'{amount} is not a valid float.')
        state.require(
            float(percent_value) >= 0 and float(percent_value) <= 100, 
            f'{amount} is not a valid agent amount.', 
            'Amount must be non-negative.'
        )

        agent_amount = AgentPercentAmount(percent_value)

    else:
        state.require(is_int(amount), f'{amount} is not a valid integer.')
        state.require(int(amount) >= 0, f'{amount} is not a valid agent amount.', 'Amount must be non-negative.')

        agent_amount = AgentConstantAmount(amount)
            

    connection_amount: ConnectionAmount | None = None
    match args:
        case [ value ]:
            state.require(is_int(value), f'{value} is not a valid integer.')
            state.require(
                int(value) >= 0, 
                f'{value} is not a valid connection amount.', 
                'Amount must be non-negative.'
            )

            connection_amount = ConnectionConstantAmount(value)

        case [ 'dist_normal', mean, std_dev ]:
            state.require(is_float(mean), f'{mean} is not a valid float.')
            state.require(is_float(std_dev), f'{std_dev} is not a valid float.')
            state.require(
                float(std_dev) >= 0, 
                f'{std_dev} is not a valid standard deviation parameter.', 
                'Standard deviation must be non-negative.'
            )

            connection_amount = ConnectionDistNormalAmount(mean, std_dev)
            
        case [ 'dist_exp', lambda_ ]:
            state.require(is_float(lambda_), f'{lambda_} is not a valid float.')
            state.require(
                float(lambda_) > 0, 
                f'{lambda_} is not a valid lambda parameter.', 
                'Lambda must be positive.'
            )
            
            connection_amount = ConnectionDistExpAmount(lambda_)
            
        case [ 'dist_uniform', a, b ]:
            state.require(is_float(a), f'{a} is not a valid float.')
            state.require(is_float(b), f'{b} is not a valid float.')
            
            connection_amount = ConnectionDistUniformAmount(a, b)

        case _:
            state.panic(f'Incorrect operation: DEFG {agent_name} {amount} {args}')

    state.last_graph.add_agent(StatisticalAgent(agent_name, agent_amount, connection_amount))



from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple, TypeVar

if TYPE_CHECKING:
    T = TypeVar("T")


def zip_consecutive_pairs(iterable: List[T]) -> List[Tuple[T, ...]]:
    if len(iterable) % 2 != 0:
        raise Exception("Iterable must have an even number of elements")
    iterator = iter(iterable)
    double_iterator = [iterator] * 2
    return list(zip(*double_iterator))



from __future__ import annotations

from typing import Dict, List


class FloatParam:
    def __init__(self, name: str):
        self.name: str = name
        self.value: str = ''
        self.is_value_set: bool = False
        
    def set_value(self, value: str) -> None:
        self.is_value_set = True
        self.value = value
        
    def print(self) -> None:
        print(f'MessageFloatParam {self.name} = {self.value}')


class Message:
    RESERVED_CONNECTION_PARAMS = [ 'sender' ]
    RESERVED_TYPE_PARAMS = [ 'type', 'performative' ]
    
    def __init__(self, msg_type: str, msg_performative: str):
        self.type: str = msg_type
        self.performative: str = msg_performative
        self.float_params: Dict[str, FloatParam] = {}
        
    @property
    def param_names(self) -> List[str]:
        return [ *Message.RESERVED_CONNECTION_PARAMS,
                 *Message.RESERVED_TYPE_PARAMS,
                 *list(self.float_params) ]
    
    @property
    def unset_params(self) -> List[str]:
        unset_params: List[str] = []
        for name, float_param in self.float_params.items():
            if not float_param.is_value_set:
                unset_params.append(name)
        return unset_params
    
    def are_all_params_set(self) -> bool:
        return all([float_param.is_value_set for float_param in self.float_params.values()])
    
    def param_exists(self, name: str) -> bool:
        return name in self.param_names
    
    def add_float(self, float_param: FloatParam) -> None:
        self.float_params[float_param.name] = float_param
        
    def print(self) -> None:
        print(f'Message {self.type}/{self.performative}')
        for float_params in self.float_params.values():
            float_params.print()
